import logging
import os
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import TypeVar

import jsonpointer

from qualibrate_config.cli import migrate_command
from qualibrate_config.file import get_config_file, read_config_file
from qualibrate_config.models import BaseConfig, QualibrateConfig
from qualibrate_config.models.qualibrate import QualibrateTopLevelConfig
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.validation import (
    GreaterThanSupportedQualibrateConfigVersionError,
    InvalidQualibrateConfigVersionError,
    get_config_model_or_print_error,
    qualibrate_version_validator,
)
from qualibrate_config.vars import (
    CONFIG_PATH_ENV_NAME,
    DEFAULT_CONFIG_FILENAME,
    QUALIBRATE_PATH,
)

__all__ = [
    "get_config_dict",
    "get_config_model",
    "get_qualibrate_config_path",
    "get_qualibrate_config",
]

ConfigClass = TypeVar("ConfigClass", bound=BaseConfig)


def get_qualibrate_config_path() -> Path:
    """
    Retrieve the qualibrate configuration file path. If an environment variable
    for the config path is set, it uses that; otherwise, it defaults to the
    standard Qualibrate path.
    """
    return get_config_file(
        os.environ.get(CONFIG_PATH_ENV_NAME, QUALIBRATE_PATH),
        DEFAULT_CONFIG_FILENAME,
        raise_not_exists=False,
    )


def get_config_dict(
    config_path: Path,
    config_key: str | None,
    config: RawConfigType | None = None,
) -> RawConfigType:
    if config is None or config_key is None or config_key not in config:
        config = read_config_file(config_path, solve_references=False)
    if config is None:
        raise RuntimeError("Couldn't read config file")
    if config_key is None:
        return dict(config)
    return dict(config.get(config_key, {}))


def get_config_dict_from_subpath(
    config_path: Path,
    subpath: str | None,
    config: RawConfigType | None = None,
) -> RawConfigType:
    """
    Retrieves a configuration dictionary and optionally resolves a subpath
    within it.

    Args:
        config_path: The path to the configuration file.
        subpath: The optional JSON pointer string indicating the subpath
            within the configuration dictionary to retrieve. If None, the
            entire configuration dictionary is returned.
        config: An optional existing configuration dictionary to use instead
            of reading from the file. Defaults to None.

    Returns:
        A dictionary representing the configuration or the resolved subpath
        within it.
    """
    config_dict = get_config_dict(config_path, None, config)
    if subpath is None:
        return config_dict
    return dict(jsonpointer.resolve_pointer(config_dict, subpath))


def get_config_model(
    config_path: Path,
    config_key: str | None,
    config_class: type[ConfigClass] = QualibrateConfig,  # type: ignore
    config: RawConfigType | None = None,
    raw_config_validators: list[Callable[[RawConfigType], None]] | None = None,
) -> ConfigClass:
    """Retrieve the configuration settings.

        Args:
        config_path: Path to the configuration file.
        config: Optional pre-loaded configuration data. If not provided, it
            will load and resolve references from the config file.

    Returns:
        An instance of QualibrateConfig with the loaded configuration.

    Raises:
        RuntimeError: If the configuration file cannot be read or if the
            configuration state is invalid.
    """
    model_config_dict = get_config_dict(config_path, config_key, config)
    if raw_config_validators:
        for validator in raw_config_validators:
            validator(model_config_dict)
    new_config = get_config_model_or_print_error(
        model_config_dict,
        config_class,
        config_key,
    )
    if new_config is None:
        raise RuntimeError(f"Invalid config {config_class.__name__} state")
    return new_config


def get_qualibrate_config(
    config_path: Path | None = None,
    config: RawConfigType | None = None,
    auto_migrate: bool = True,
) -> QualibrateConfig:
    """Retrieve the Qualibrate configuration.

    Args:
        config_path: Optional pre-loaded configuration data. If not provided, it
            will load and resolve references from the config file.
        config: Optional pre-loaded configuration data. If not provided, it
            will load and resolve references from the config file.
        auto_migrate: is it needed to automatically apply migrations to config

    Returns:
        An instance of QualibrateConfig with the loaded configuration.

    Raises:
        RuntimeError: If the configuration file cannot be read or if the
            configuration state is invalid.
    """
    if config_path is None:
        config_path = get_qualibrate_config_path()

    get_config_model_part = partial(
        get_config_model,
        config_path,
        config_key=None,
        config_class=QualibrateTopLevelConfig,
        config=config,
    )
    common_error_msg = (
        "QUAlibrate was unable to load the config. It is recommend to run "
        '"qualibrate config" to fix any file issues. If this problem persists, '
        f'please delete "{config_path}" and retry running '
        '"qualibrate config"'
    )
    try:
        model = get_config_model_part(
            raw_config_validators=[qualibrate_version_validator]
        )
    except GreaterThanSupportedQualibrateConfigVersionError as ex:
        error_msg = (
            f"QUAlibrate was unable to load the config. {str(ex)}. If this "
            f'problem persists, please delete "{config_path}" and retry '
            'running "qualibrate config"'
        )
        raise RuntimeError(error_msg) from ex
    except InvalidQualibrateConfigVersionError:
        if not auto_migrate:
            raise
        logging.info("Automatically migrate to new qualibrate config")
        migrate_command(["--config-path", config_path], standalone_mode=False)
    except (RuntimeError, ValueError) as ex:
        raise RuntimeError(common_error_msg) from ex
    else:
        return model.qualibrate
    # migrated
    try:
        model = get_config_model_part()
    except (RuntimeError, ValueError) as ex:
        raise RuntimeError(common_error_msg) from ex
    return model.qualibrate


if __name__ == "__main__":
    get_qualibrate_config(get_qualibrate_config_path())



import sys
from pathlib import Path
from typing import cast

from qualibrate_config.core.project.path import get_project_config_path
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.vars import (
    QUALIBRATE_CONFIG_KEY,
)

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib  # type: ignore[unused-ignore,import-not-found]
else:
    import tomllib


def get_project_from_common_config(config: RawConfigType) -> str | None:
    return cast(
        str | None, config.get(QUALIBRATE_CONFIG_KEY, {}).get("project")
    )


def read_project_config_file(
    config_path: Path,
    project_name: str,
) -> RawConfigType:
    project_config = get_project_config_path(config_path.parent, project_name)
    if not project_config.is_file():
        return {}
    with project_config.open("rb") as fin:
        return tomllib.load(fin)



from pathlib import Path

import click

from qualibrate_config.core.content import get_config_file_content, simple_write
from qualibrate_config.core.project.path import get_project_path
from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY


def switch_project(
    config_path: Path, project: str, *, raise_if_error: bool = False
) -> bool:
    project_path = get_project_path(config_path.parent, project)
    if not project_path.exists():
        error_msg = (
            f"Can't switch project to '{project}'. "
            f"There is no specified project in {project_path.parent}."
        )
        if raise_if_error:
            raise ValueError(error_msg)
        click.secho(error_msg, fg="red")
        return False
    common_config, config_file = get_config_file_content(config_path)
    common_config[QUALIBRATE_CONFIG_KEY]["project"] = project
    simple_write(config_file, common_config)
    return True



from .create import create_command
from .current import current_command
from .delete import delete_command
from .list import list_command
from .switch import switch_command

__all__ = [
    "create_command",
    "current_command",
    "delete_command",
    "list_command",
    "switch_command",
]



from functools import cached_property

from qualibrate_config.models.base.config_base import BaseConfig

__all__ = [
    "RemoteServiceBaseConfig",
    "SpawnServiceBaseConfig",
    "JsonTimelineDBRemoteServiceConfig",
    "QualibrateRunnerRemoteServiceConfig",
    "QualibrateAppSubServiceConfig",
    "QualibrateRunnerSubServiceConfig",
    "QuaDashboardSubServiceConfig",
]


class RemoteServiceBaseConfig(BaseConfig):
    address: str
    timeout: float

    @cached_property
    def address_with_root(self) -> str:
        address = str(self.address)
        return address if address.endswith("/") else address + "/"


class SpawnServiceBaseConfig(BaseConfig):
    spawn: bool


class JsonTimelineDBRemoteServiceConfig(RemoteServiceBaseConfig):
    pass


class QualibrateRunnerRemoteServiceConfig(RemoteServiceBaseConfig):
    pass


class QualibrateAppSubServiceConfig(SpawnServiceBaseConfig):
    pass


class QualibrateRunnerSubServiceConfig(SpawnServiceBaseConfig):
    pass


class QuaDashboardSubServiceConfig(SpawnServiceBaseConfig):
    pass



from pathlib import Path

import click

from qualibrate_config.cli.vars import CONFIG_PATH_HELP
from qualibrate_config.core.project.p_list import (
    print_simple_projects_list,
    print_verbose_projects_list,
)
from qualibrate_config.vars import DEFAULT_CONFIG_FILEPATH

__all__ = ["list_command"]


@click.command(name="list")
@click.option(
    "--config-path",
    type=click.Path(exists=True, path_type=Path),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
@click.option("--verbose", "-v", is_flag=True)
def list_command(config_path: Path, verbose: bool) -> None:
    print_ = (
        print_verbose_projects_list if verbose else print_simple_projects_list
    )
    print_(config_path)


if __name__ == "__main__":
    list_command([], standalone_mode=False)



import shutil
from pathlib import Path

from qualibrate_config.core.content import get_config_file_content
from qualibrate_config.core.project.common import get_project_from_common_config
from qualibrate_config.core.project.p_list import list_projects
from qualibrate_config.core.project.path import get_project_path


def delete_project(config_path: Path, project: str) -> None:
    qualibrate_path = config_path.parent
    try:
        projects = list_projects(qualibrate_path)
    except NotADirectoryError as e:
        # there is no any project
        raise RuntimeError("Projects dir doesn't exist.") from e
    if project not in projects:
        raise RuntimeError(f"Project '{project}' does not exist")
    common_config, config_file = get_config_file_content(config_path)
    current_project = get_project_from_common_config(common_config)
    if current_project and current_project == project:
        raise RuntimeError("Can't delete current project.")
    project_path = get_project_path(qualibrate_path, project)
    if project_path is None:
        raise RuntimeError(f"Can't resolve project '{project}' directory")
    shutil.rmtree(project_path)



import inspect
import sys
from collections.abc import Callable
from pathlib import Path
from typing import TypeVar, cast, overload

import tomli_w

from qualibrate_config.core.approve import print_and_confirm
from qualibrate_config.core.project.path import get_project_path
from qualibrate_config.file import get_config_file
from qualibrate_config.models import BaseConfig, QualibrateConfig
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.vars import DEFAULT_CONFIG_FILENAME

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib  # type: ignore[unused-ignore,import-not-found]
else:
    import tomllib


__all__ = [
    "ConfigType",
    "get_config_file_content",
    "simple_write",
    "qualibrate_after_write_cb",
    "write_config",
]

ConfigType = TypeVar("ConfigType", bound=BaseConfig)
# Non-generic callback alias to avoid "Missing type parameters" errors
Callback = (
    Callable[[BaseConfig], None] | Callable[[BaseConfig, Path | None], None]
)


def get_config_file_content(config_path: Path) -> tuple[RawConfigType, Path]:
    """Returns config and path to file"""
    config_file = get_config_file(
        config_path, DEFAULT_CONFIG_FILENAME, raise_not_exists=False
    )
    # TODO: first location of tomllib.loads
    if config_file.is_file():
        return tomllib.loads(config_file.read_text()), config_path
    return {}, config_file


def simple_write(path: Path, config: RawConfigType) -> None:
    with path.open("wb") as f_out:
        tomli_w.dump(config, f_out)


def qualibrate_after_write_cb(
    config: BaseConfig,
    config_file: Path | None = None,
) -> None:
    if config.project in config.storage.location.parts:
        config.storage.location.mkdir(parents=True, exist_ok=True)
    else:
        (config.storage.location / config.project).mkdir(
            parents=True, exist_ok=True
        )
    if config_file is not None:
        project_path = get_project_path(config_file.parent, config.project)
        project_path.mkdir(parents=True, exist_ok=True)
        (project_path / DEFAULT_CONFIG_FILENAME).touch()


# Overloads tell mypy the two legal shapes
@overload
def _call_cb(cb: None, config: BaseConfig, config_file: Path) -> None: ...
@overload
def _call_cb(
    cb: Callable[[BaseConfig], None], config: BaseConfig, config_file: Path
) -> None: ...
@overload
def _call_cb(
    cb: Callable[[BaseConfig, Path | None], None],
    config: BaseConfig,
    config_file: Path,
) -> None: ...
def _call_cb(
    cb: Callback | None,
    config: BaseConfig,
    config_file: Path,
) -> None:
    if cb is None:
        return
    # Runtime dispatch, then cast to keep mypy satisfied
    params_cnt = len(inspect.signature(cb).parameters)
    if params_cnt == 1:
        cb1 = cast(Callable[[BaseConfig], None], cb)
        cb1(config)
    else:
        cb2 = cast(Callable[[BaseConfig, Path | None], None], cb)
        cb2(config, config_file)


def write_config(
    config_file: Path,
    common_config: RawConfigType,
    config: ConfigType,
    config_key: str,
    before_write_cb: Callback | None = None,
    after_write_cb: Callback | None = None,
    confirm: bool = True,
    check_generator: bool = False,
) -> None:
    exported_data = config.serialize(exclude_none=True)
    common_config[config_key] = exported_data
    if confirm or check_generator:
        print_and_confirm(config_file, common_config, check_generator)

    _call_cb(before_write_cb, config, config_file)

    if not config_file.parent.exists():
        config_file.parent.mkdir(parents=True)
    simple_write(config_file, common_config)

    if after_write_cb is None and isinstance(config, QualibrateConfig):
        after_write_cb = qualibrate_after_write_cb
    _call_cb(after_write_cb, config, config_file)



from pathlib import Path

from qualibrate_config.models.base.config_base import BaseConfig
from qualibrate_config.models.calibration_library import (
    CalibrationLibraryConfig,
)
from qualibrate_config.models.composite import QualibrateCompositeConfig
from qualibrate_config.models.q_app import QualibrateAppConfig
from qualibrate_config.models.remote_services import (
    QualibrateRunnerRemoteServiceConfig,
)
from qualibrate_config.models.storage import (
    StorageConfig,
)

__all__ = ["QualibrateConfig", "QualibrateTopLevelConfig"]


class QualibrateConfig(BaseConfig):
    version: int = 5
    project: str = "init_project"
    password: str | None = None
    log_folder: Path | None = None

    storage: StorageConfig
    app: QualibrateAppConfig | None = None
    runner: QualibrateRunnerRemoteServiceConfig | None = None
    composite: QualibrateCompositeConfig | None = None
    calibration_library: CalibrationLibraryConfig | None = None


class QualibrateTopLevelConfig(BaseConfig):
    qualibrate: QualibrateConfig



import sys
from pathlib import Path

from qualibrate_config.core.project.common import (
    get_project_from_common_config,
    read_project_config_file,
)
from qualibrate_config.core.utils import recursive_update_dict
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.references.resolvers import resolve_references
from qualibrate_config.vars import (
    DEFAULT_CONFIG_FILENAME,
    QUALIBRATE_CONFIG_KEY,
    QUALIBRATE_PATH,
)

if sys.version_info[:2] < (3, 11):
    import tomli as tomllib  # type: ignore[unused-ignore,import-not-found]
else:
    import tomllib


def _get_config_file_from_dir(
    dir_path: Path,
    default_config_specific_filename: str | Path,
    raise_not_exists: bool = True,
) -> Path:
    default_qualibrate = dir_path / default_config_specific_filename
    if default_qualibrate.is_file():
        return default_qualibrate
    default_common = dir_path / DEFAULT_CONFIG_FILENAME
    if default_common.is_file():
        return default_common
    if raise_not_exists:
        raise FileNotFoundError(f"Config file in dir {dir_path} does not exist")
    return default_common


def get_config_file(
    config_path: str | Path | None,
    default_config_specific_filename: str | Path,
    raise_not_exists: bool = True,
) -> Path:
    if config_path is None:
        return _get_config_file_from_dir(
            QUALIBRATE_PATH, default_config_specific_filename, raise_not_exists
        )
    config_path_ = Path(config_path)
    if config_path_.is_file():
        return config_path_
    if config_path_.is_dir():
        return _get_config_file_from_dir(
            config_path_, default_config_specific_filename
        )
    if raise_not_exists:
        raise OSError("Unexpected config file path")
    return config_path_


def read_config_file(
    config_file: Path,
    solve_references: bool = True,
    override_project: str | None = None,
) -> RawConfigType:
    # TODO: second location of tomllib.loads
    with config_file.open("rb") as fin:
        config: RawConfigType = tomllib.load(fin)  # typing for mypy tomli
    if project := (override_project or get_project_from_common_config(config)):
        project_config = read_project_config_file(config_file, project)
        project_config.setdefault(QUALIBRATE_CONFIG_KEY, {})["project"] = (
            project
        )
        config = recursive_update_dict(config, project_config)
    if not solve_references:
        return config
    return resolve_references(config)



from collections.abc import Mapping
from io import StringIO
from typing import Annotated, Any

from pydantic import AwareDatetime, BaseModel, Field, field_serializer

from qualibrate_config.core.approve import print_config


class Project(BaseModel):
    name: str
    nodes_number: int
    created_at: AwareDatetime
    last_modified_at: AwareDatetime
    updates: Annotated[Mapping[str, Any], Field(default_factory=dict)]

    @field_serializer("created_at", "last_modified_at")
    def dt_serializer(self, dt: AwareDatetime) -> str:
        return dt.isoformat(timespec="seconds")

    def verbose_str(self) -> str:
        d = self.model_dump()
        d.pop("name")
        updates = d.pop("updates")
        if updates:
            updates_io = StringIO()
            print_config(updates, 2, updates_io)
            updates_str = "config overrides:\n" + updates_io.getvalue()
        else:
            updates_str = "config overrides: None"
        return f"\n{' ' * 4}".join(
            [
                f"Project: {self.name}",
                *[f"{k}: {v}" for k, v in d.items()],
                updates_str,
            ]
        ).rstrip("\n")



from importlib.util import find_spec
from pathlib import Path
from typing import Annotated

from qualibrate_config.models.base.config_base import BaseConfig
from qualibrate_config.models.base.default_value import DefaultConfigValue
from qualibrate_config.models.remote_services import (
    JsonTimelineDBRemoteServiceConfig,
)

__all__ = ["QualibrateAppConfig"]


def get_default_static_path() -> Path | None:
    app_module = find_spec("qualibrate_app")
    if app_module is None or app_module.origin is None:
        return None
    return Path(app_module.origin).resolve().parents[1] / "qualibrate_static"


class QualibrateAppConfig(BaseConfig):
    static_site_files: Annotated[
        Path | None, DefaultConfigValue(factory=get_default_static_path)
    ] = None

    timeline_db: JsonTimelineDBRemoteServiceConfig | None = None



CONFIG_PATH_HELP = (
    "Path to the configuration file. If the path points to a file, it will "
    "be read and the old configuration will be reused, except for the "
    "variables specified by the user. If the file does not exist, a new one"
    " will be created. If the path points to a directory, a check will be "
    "made to see if files with the default name exist."
)
CALIBRATION_LIBRARY_FOLDER_HELP = (
    "Path to the folder contains calibration nodes and graphs."
)
STORAGE_LOCATION_HELP = (
    "Path to the local user storage. Used for storing nodes output data."
)
QUAM_STATE_PATH_HELP = "Path to the quam state."



import sys
from pathlib import Path

import click

from qualibrate_config.cli.vars import CONFIG_PATH_HELP
from qualibrate_config.core.project.switch import switch_project
from qualibrate_config.vars import DEFAULT_CONFIG_FILEPATH

__all__ = ["switch_command"]


@click.command(name="switch")
@click.argument("project", type=str)
@click.option(
    "--config-path",
    type=click.Path(exists=True, path_type=Path),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
def switch_command(project: str, config_path: Path) -> None:
    if switch_project(config_path, project):
        click.echo(f"Project switched to '{project}'.")
    else:
        sys.exit(1)


if __name__ == "__main__":
    switch_command([], standalone_mode=False)



import functools
import warnings

from qualibrate_config.cli.utils import deprecated_class_alias
from qualibrate_config.core.migration.migrations import base as _base_content

deprecated_m = "qualibrate_config.cli.migrations.base"
new_m = "qualibrate_config.core.migration.migrations.base"

warnings.warn(
    (
        f"Module '{deprecated_m}' is deprecated and will be removed in a "
        f"future version. Please use '{new_m}' instead."
    ),
    DeprecationWarning,
    stacklevel=2,
)

cl_deprecated = functools.partial(
    deprecated_class_alias, deprecated_module=deprecated_m, new_module=new_m
)
# Apply decorator to each re-exported function
MigrateBase = cl_deprecated(name="MigrateBase")(_base_content.MigrateBase)



import importlib
import types
import typing
from collections.abc import Callable
from enum import Enum
from itertools import filterfalse
from pathlib import Path
from typing import (
    Annotated,
    Any,
    ClassVar,
    Optional,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic_core._pydantic_core import ValidationError

from qualibrate_config.models.base.default_value import DefaultConfigValue
from qualibrate_config.models.base.importable import Importable
from qualibrate_config.models.base.path_serializer import PathSerializer
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.references.resolvers import (
    TEMPLATE_START,
    resolve_single_item,
)

__all__ = ["BaseConfig"]


class BaseConfig:
    """
    A base class for handling nested configuration with type annotations,
    default values, optional fields, and attribute-style access.
    """

    _root: ClassVar[Optional["BaseConfig"]] = None

    def __init__(
        self,
        config: RawConfigType,
        path: str | None = None,
        root: Optional["BaseConfig"] = None,
    ) -> None:
        """
        Initializes the configuration object with the provided dictionary.
        Supports nested configurations, default values, and optional fields.
        """
        self._path = path or "/"
        self._data: RawConfigType = {}
        self._annotations = self.get_config_annotations()
        self._raw_dict: RawConfigType = {}
        self.__class__._root = root or self

        for key, value_type_default in self._annotations.items():
            value_type = (
                value_type_default[0]
                if isinstance(value_type_default, tuple)
                else value_type_default
            )
            # Default value from class or set to None
            default_value = getattr(self.__class__, key, None)
            # Get value from config or use default
            value = config.get(key, default_value)

            value = self._parse_value(key, value, value_type)
            self._set_config_attr(key, value)
        for key in filterfalse(self._annotations.__contains__, config):
            self._raw_dict[key] = config[key]

    def _set_config_attr(self, key: str, value: Any) -> None:
        raw_value = value._raw_dict if isinstance(value, BaseConfig) else value
        self._raw_dict[key] = raw_value
        self._data[key] = value

    def _is_reference_key(self, attr_name: str) -> bool:
        if attr_name not in self._annotations:
            return False
        attr_value = self._data[attr_name]
        return self._is_reference(attr_value)

    @staticmethod
    def _is_reference(attr_value: Any) -> bool:
        return isinstance(attr_value, str) and TEMPLATE_START in attr_value

    def _get_referenced_value(self, attr_name: str) -> Any:
        if attr_name not in self._annotations:
            raise AttributeError(
                f"{self.__class__.__name__} has no attribute {attr_name}"
            )
        reference = self._data.get(attr_name)
        if reference is None:
            raise ValueError(f"Attribute {attr_name} has no reference")
        return resolve_single_item(self._get_root()._data, reference)

    @classmethod
    def _get_root(cls) -> "BaseConfig":
        if cls._root is None:
            raise RuntimeError("Root can't be None")
        return cls._root

    def _parse_type_origin(
        self, key: str, value: Any, expected_type: type
    ) -> Any:
        error_path = ("/", *filter(bool, self._path.split("/")), key)
        if issubclass(expected_type, list):
            if not isinstance(value, list):
                from pydantic_core import InitErrorDetails

                raise ValidationError.from_exception_data(
                    f"value must be list, got {type(value).__name__}.",
                    line_errors=[
                        InitErrorDetails(
                            type="InvalidType",
                            loc=error_path,
                            input=value,
                        )
                    ],
                )
            # TODO: check args type
            return value

        if issubclass(expected_type, dict):
            # Handle dict types
            if not isinstance(value, dict):
                raise ValueError(
                    f"Field '{key}' expects a dict, got {type(value).__name__}."
                )
            # TODO: check args type
            return value

        if issubclass(expected_type, BaseConfig):
            # Handle nested configuration
            path = f"{self._path.rstrip('/')}/{key}"
            if isinstance(value, expected_type):
                value._path = path
                return value
            if isinstance(value, dict):
                return expected_type(value, path=path, root=self._get_root())
            raise ValueError(
                f"Field '{key}' expects a dictionary for nested config, "
                f"got {type(value).__name__}."
            )

        if issubclass(expected_type, Enum):
            return expected_type(value)
        if issubclass(expected_type, Path):
            return Path(value)

        if issubclass(expected_type, Importable) and isinstance(value, str):
            return value
        return None

    def _parse_value(self, key: str, value: Any, expected_type: type) -> Any:
        """Parses and validates a value based on the expected type."""
        origin = get_origin(expected_type) or expected_type
        args = get_args(expected_type)

        if origin is types.UnionType or origin is typing.Union:
            # Handle Union types, including Optional (Union[..., None])
            if type(None) in args and value is None:
                return value
            exs: list[Exception] = []
            for arg in args:
                try:
                    return self._parse_value(key, value, arg)
                except (TypeError, ValueError) as ex:
                    exs.append(ex)
            raise ValueError(
                f"Field '{key}' expects one of {args}, "
                f"got {type(value).__name__}."
            ) from RuntimeError(exs)

        if isinstance(origin, type):
            result = self._parse_type_origin(key, value, origin)
            if result is not None:
                return result
        if self._is_reference(value):
            return value

        # Validate primitive types
        if issubclass(origin, float) and isinstance(value, int):
            return origin(value)
        if not isinstance(value, origin):
            raise ValueError(
                f"Field '{key}' expects {expected_type.__name__}, "
                f"got {type(value).__name__}."
            )
        return value

    @classmethod
    def get_config_annotations(
        cls,
    ) -> dict[str, tuple[type, DefaultConfigValue]]:
        """
        Collect annotations from the current class and its ancestors.
        Handles inheritance correctly.
        """
        annotations = {}
        for parent in reversed(cls.__mro__):
            if issubclass(parent, BaseConfig):
                type_hints = get_type_hints(parent)
                class_annotations = parent.__annotations__
                for attr, annot in class_annotations.items():
                    if attr not in type_hints:
                        raise AttributeError("Unknown attribute")
                    if get_origin(annot) is not Annotated:  # type: ignore[comparison-overlap]
                        type_hints[attr] = (type_hints[attr], None)
                        continue
                    args = get_args(annot)
                    annot_default = next(
                        filter(
                            lambda a: isinstance(a, DefaultConfigValue), args
                        ),
                        None,
                    )
                    if annot_default is not None:
                        type_hints[attr] = (type_hints[attr], annot_default)
                annotations.update(type_hints)
        annotations.pop("_root", None)
        return annotations

    def serialize(self, exclude_none: bool = True) -> RawConfigType:
        def _get_val(val: Any) -> Any:
            if isinstance(val, BaseConfig):
                return val.serialize()
            if isinstance(val, Path):
                return PathSerializer.serialize_path(val)
            if isinstance(val, Enum):
                return val.value
            return val

        conditions: list[Callable[[str, Any], bool]] = []
        if exclude_none:
            conditions.append(lambda k, v: v is not None)
        return {
            key: _get_val(value)
            for key, value in self._data.items()
            if (
                len(conditions) == 0
                or all(condition(key, value) for condition in conditions)
            )
        }

    def __getattribute__(self, name: str) -> Any:
        """
        Custom getter for attributes.

        :param name: Attribute name to access.
        :return: Attribute value or raises AttributeError if not found.
        """
        if name in ConfigBaseDir:
            res = super().__getattribute__(name)
            return res
        annotations = self._annotations
        if name not in annotations:
            return super().__getattribute__(name)
        data = self._data
        if name not in data:
            raise AttributeError(
                f"There is no {name} in {self.__class__.__name__}."
            )
        value = data[name]
        if self._is_reference(value):
            if self.__class__._root is None:
                raise ValueError("Root shouldn't be None")
            raw_dict = self._get_root()._raw_dict
            return resolve_single_item(raw_dict, value)
        annotation_with_default = annotations[name]
        default: DefaultConfigValue | None = None
        if isinstance(annotation_with_default, tuple):
            annotation, default = annotation_with_default
        else:
            annotation = annotation_with_default
        if value is None and isinstance(default, DefaultConfigValue):
            value = default.value
        # left or - for type annotations; right or - for types directly
        annotation_type = get_origin(annotation) or annotation
        # handle Optional
        if (
            annotation_type is Union
            and (ann_args := get_args(annotation))
            and (len(ann_args) == 2)
            and (ann_args[1] is type(None))
        ):
            annotation_type = ann_args[0]

        if not isinstance(annotation_type, type):
            return value
        if issubclass(annotation_type, Importable):
            module, class_ = value.rsplit(".", maxsplit=1)
            class_module = importlib.import_module(module)
            return getattr(class_module, class_)
        if issubclass(annotation_type, Path):
            path_str = PathSerializer.serialize_path(value)
            if not self._is_reference(path_str):
                return value
            raw_dict = self._get_root()._raw_dict
            resolved_path = resolve_single_item(raw_dict, path_str)
            return Path(resolved_path)
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Custom setter for attributes.

        :param name: Attribute name to set.
        :param value: Value to assign to the attribute.
        """
        if name in ConfigBaseDir:
            super().__setattr__(name, value)
            return
        if name not in self._annotations:
            super().__setattr__(name, value)
            return
        annot = self._annotations[name]
        expected_type = annot[0] if isinstance(annot, tuple) else annot
        value = self._parse_value(name, value, expected_type)
        self._set_config_attr(name, value)


ConfigBaseDir = (
    *dir(BaseConfig),
    "_path",
    "_data",
    "_annotations",
    "_raw_dict",
)



from pathlib import Path

__all__ = [
    "CONFIG_PATH_ENV_NAME",
    "QUALIBRATE_CONFIG_KEY",
    "QUALIBRATE_PATH",
    "DEFAULT_CONFIG_FILENAME",
    "DEFAULT_CONFIG_FILEPATH",
    "QUAM_CONFIG_KEY",
    "QUAM_STATE_PATH_CONFIG_KEY",
]
CONFIG_PATH_ENV_NAME = "QUALIBRATE_CONFIG_FILE"

QUALIBRATE_CONFIG_KEY = "qualibrate"
QUAM_CONFIG_KEY = "quam"
QUAM_STATE_PATH_CONFIG_KEY = "state_path"

QUALIBRATE_PATH = Path().home() / ".qualibrate"
DEFAULT_CONFIG_FILENAME = "config.toml"

DEFAULT_CONFIG_FILEPATH = QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME



import click

from .commands import (
    create_command,
    current_command,
    delete_command,
    list_command,
    switch_command,
)


@click.group(name="project", help="Manage projects")
def project_group() -> None:
    pass


project_group.add_command(create_command)
project_group.add_command(current_command)
project_group.add_command(delete_command)
project_group.add_command(list_command)
project_group.add_command(switch_command)



from collections.abc import Callable
from functools import cached_property
from typing import Any


class DefaultConfigValue:
    def __init__(
        self,
        value: Any = None,
        factory: Callable[[], Any] | None = None,
    ):
        if value and factory:
            raise ValueError("Cannot set both factory and value")
        if value is None and factory is None:
            raise ValueError("Cannot set `None` both factory and value")
        self._value = value
        self._factory = factory

    @cached_property
    def value(self) -> Any:
        if self._factory is not None and self._value is None:
            self._value = self._factory()
        return self._value



import functools
import warnings

from qualibrate_config.cli.utils import deprecated_alias
from qualibrate_config.core import content as _core_content

__all__ = _core_content.__all__

deprecated_m = "qualibrate_config.cli.utils.content"
new_m = "qualibrate_config.core.content"

warnings.warn(
    (
        f"Module '{deprecated_m}' is deprecated and will be removed in a "
        f"future version. Please use '{new_m}' instead."
    ),
    DeprecationWarning,
    stacklevel=2,
)

l_deprecated = functools.partial(
    deprecated_alias, deprecated_module=deprecated_m, new_module=new_m
)
# Apply decorator to each re-exported function
get_config_file_content = l_deprecated(
    name="get_config_file_content",
)(_core_content.get_config_file_content)
simple_write = l_deprecated(name="simple_write")(_core_content.simple_write)
write_config = l_deprecated(name="write_config")(_core_content.write_config)



import importlib
from collections.abc import Callable
from pathlib import Path
from typing import cast

from qualibrate_config.core.migration.migrations.base import MigrateBase
from qualibrate_config.qulibrate_types import RawConfigType

__all__ = ["migration_direction", "migration_functions", "make_migrations"]


def migration_direction(from_version: int, to_version: int) -> bool:
    return to_version > from_version


def migration_functions(
    from_version: int,
    to_version: int,
    migrations_package: str,
    module_name_format: str,
) -> list[
    Callable[[RawConfigType, Path], RawConfigType]
    | Callable[[RawConfigType], RawConfigType]
]:
    direction = migration_direction(from_version, to_version)
    version_step = 1 if direction else -1
    versions = range(from_version, to_version, version_step)
    functions = []
    for version in versions:
        other_version = version + version_step
        minmax = (
            (version, other_version) if direction else (other_version, version)
        )
        module_name = module_name_format.format(minmax[0], minmax[1])
        full_module_name = f"{migrations_package}.{module_name}"
        module = importlib.import_module(full_module_name)
        if not hasattr(module, "Migrate") or not issubclass(
            module.Migrate, MigrateBase
        ):
            raise AttributeError(
                f"Module {full_module_name} has no Migrate class or it "
                f"has invalid hierarchy"
            )
        function = (
            module.Migrate.forward if direction else module.Migrate.backward
        )
        functions.append(function)
    return functions


def make_migrations(
    data: RawConfigType,
    from_version: int,
    to_version: int,
    migrations_package: str = "qualibrate_config.core.migration.migrations",
    module_name_format: str = "v{}_v{}",
    *,
    config_path: Path | None = None,
) -> RawConfigType:
    functions = migration_functions(
        from_version, to_version, migrations_package, module_name_format
    )
    for function in functions:
        if function.__code__.co_argcount == 1:
            data = cast(Callable[[RawConfigType], RawConfigType], function)(
                data
            )
        else:
            if config_path is None:
                raise RuntimeError("Config path must be provided")
            data = cast(
                Callable[[RawConfigType, Path], RawConfigType], function
            )(data, config_path)
    return data



from qualibrate_config.models.base.config_base import BaseConfig
from qualibrate_config.models.remote_services import (
    QuaDashboardSubServiceConfig,
    QualibrateAppSubServiceConfig,
    QualibrateRunnerSubServiceConfig,
)

__all__ = ["QualibrateCompositeConfig"]


class QualibrateCompositeConfig(BaseConfig):
    app: QualibrateAppSubServiceConfig
    runner: QualibrateRunnerSubServiceConfig
    qua_dashboards: QuaDashboardSubServiceConfig



from pathlib import Path

import click
import tomli_w

from qualibrate_config.core.content import get_config_file_content
from qualibrate_config.core.migration.utils import make_migrations
from qualibrate_config.models import QualibrateConfig
from qualibrate_config.models.qualibrate import QualibrateTopLevelConfig
from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY

__all__ = ["run_migrations"]


def run_migrations(
    config_path: Path, to_version: int = QualibrateConfig.version
) -> None:
    common_config, config_file = get_config_file_content(config_path)
    if common_config == {}:
        click.secho("Config file wasn't found. Nothing to migrate", fg="yellow")
        return
    qualibrate_config = common_config.get(QUALIBRATE_CONFIG_KEY, {})
    from_version = qualibrate_config.get("version") or qualibrate_config.get(
        "config_version"
    )
    if from_version is None:
        click.secho(
            "Can't resolve current config version from file. Please regenerate "
            "config using `qualibrate-config config` command.",
            fg="yellow",
        )
        return
    if from_version == to_version:
        click.echo("You have latest config version. Nothing to migrate.")
        return
    if from_version > to_version:
        # TODO: merge with `qualibrate_version_validator`
        click.secho(
            (
                f"You have config version ({from_version}) greater than "
                f"supported ({QualibrateConfig.version}). Please update your "
                f"qualibrate-config package (pip install --upgrade "
                f"qualibrate-config)."
            ),
            fg="yellow",
        )
        return
    migrated = make_migrations(
        common_config, from_version, to_version, config_path=config_file
    )
    if to_version == QualibrateConfig.version:
        QualibrateTopLevelConfig(migrated)
    with config_file.open("wb") as f_out:
        tomli_w.dump(migrated, f_out)



from pathlib import Path

from qualibrate_config.core.content import get_config_file_content
from qualibrate_config.core.project.common import get_project_from_common_config
from qualibrate_config.core.project.path import get_project_config_path


def get_active_project(config_path: Path) -> str | None:
    common_config, config_file = get_config_file_content(config_path)
    project_name = get_project_from_common_config(common_config)
    if project_name is None:
        return None
    project_config = get_project_config_path(config_path.parent, project_name)
    if not project_config.is_file():
        return None
    return project_name



import functools
import warnings
from collections.abc import Callable
from typing import Any, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])
C = TypeVar("C", bound=type[Any])


def deprecated_alias(
    *, name: str, deprecated_module: str, new_module: str
) -> Callable[[F], F]:
    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(
                f"'{name}' from '{deprecated_module}.{name}' "
                f"is deprecated. Use '{new_module}.{name}' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped  # type: ignore

    return decorator


def deprecated_class_alias(
    *, name: str, deprecated_module: str, new_module: str
) -> Callable[[C], C]:
    """Decorator for classes."""

    def decorator(cls: C) -> C:
        class Wrapped(cls):  # type: ignore[misc, valid-type]
            def __init__(self, *args: Any, **kwargs: Any) -> None:
                warnings.warn(
                    f"'{name}' from '{deprecated_module}.{name}' "
                    f"is deprecated. Use '{new_module}.{name}' instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                super().__init__(*args, **kwargs)

        Wrapped.__name__ = cls.__name__
        Wrapped.__doc__ = cls.__doc__
        Wrapped.__module__ = cls.__module__
        return cast(C, Wrapped)

    return decorator



from copy import deepcopy
from pathlib import Path

from qualibrate_config.core.migration.migrations.base import MigrateBase
from qualibrate_config.qulibrate_types import RawConfigType


class Migrate(MigrateBase):
    from_version: int = 2
    to_version: int = 3

    @staticmethod
    def backward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        qualibrate = data.pop("qualibrate")
        assert qualibrate.pop("version") == Migrate.to_version
        qualibrate["version"] = Migrate.from_version

        composite: RawConfigType | None = qualibrate.get("composite", None)
        if composite:
            composite.pop("qua_dashboards", None)

        return {
            "qualibrate": qualibrate,
            **data,
        }

    @staticmethod
    def forward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        new_qualibrate = data.pop("qualibrate")
        version = new_qualibrate.pop("version")
        assert version == Migrate.from_version
        if composite := new_qualibrate.get("composite", None):
            composite.update({"qua_dashboards": {"spawn": True}})
        new_qualibrate["version"] = Migrate.to_version
        new_data = {"qualibrate": new_qualibrate, **data}
        return new_data



from pathlib import Path
from typing import (
    Any,
    TypeVar,
)

import click
from pydantic import ValidationError

from qualibrate_config.core.content import get_config_file_content
from qualibrate_config.core.migration.migrate import run_migrations
from qualibrate_config.file import read_config_file
from qualibrate_config.models import BaseConfig, QualibrateConfig
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY

T = TypeVar("T", bound=BaseConfig)

SUGGEST_MSG = (
    "Can't parse existing config. Fix it or overwrite "
    "by default values using `--overwrite` flag."
)


class WriteExitStatus(RuntimeError):
    def __init__(self, *args: Any, exit_code: int) -> None:
        super().__init__(*args)
        self.exit_code = exit_code


class InvalidQualibrateConfigVersion(RuntimeError):
    pass


class InvalidQualibrateConfigVersionError(InvalidQualibrateConfigVersion):
    def __init__(
        self,
        *args: Any,
        passed: Any = None,
        supported: int,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._passed_version = passed
        self._supported_version = supported

    def __str__(self) -> str:
        return (
            f"{super().__str__()} "
            f"Passed version: {self._passed_version}. "
            f"Supported version: {self._supported_version}"
        )

    @property
    def passed_version(self) -> Any:
        return self._passed_version

    @property
    def supported_version(self) -> int:
        return self._supported_version


class GreaterThanSupportedQualibrateConfigVersionError(
    InvalidQualibrateConfigVersionError
):
    pass


def qualibrate_version_validator(
    config: RawConfigType,
    skip_if_none: bool = True,
) -> None:
    if QUALIBRATE_CONFIG_KEY not in config:
        if skip_if_none:
            return
        raise InvalidQualibrateConfigVersionError(
            "Qualibrate config has no 'qualibrate' key",
            supported=QualibrateConfig.version,
        )
    version = config[QUALIBRATE_CONFIG_KEY].get("version")
    error_msg = (
        "QUAlibrate was unable to load the config. Can't parse version "
        "of qualibrate config. Please run `qualibrate-config config`."
    )
    if version is None:
        if skip_if_none:
            return
        raise InvalidQualibrateConfigVersionError(
            error_msg,
            passed=version,
            supported=QualibrateConfig.version,
        )
    if not isinstance(version, int):
        raise InvalidQualibrateConfigVersionError(
            error_msg,
            passed=version,
            supported=QualibrateConfig.version,
        )
    if version == QualibrateConfig.version:
        return
    if version < QualibrateConfig.version:
        raise InvalidQualibrateConfigVersionError(
            (
                "You have old version of config. "
                "Please run `qualibrate-config migrate`."
            ),
            passed=version,
            supported=QualibrateConfig.version,
        )
    raise GreaterThanSupportedQualibrateConfigVersionError(
        (
            "You have config version greater than supported. "
            "Please update your qualibrate-config package "
            "(pip install --upgrade qualibrate-config)."
        ),
        passed=version,
        supported=QualibrateConfig.version,
    )


def validate_version_and_migrate_if_needed(
    common_config: dict[str, Any],
    config_path: Path,
) -> tuple[dict[str, Any], Path]:
    try:
        qualibrate_version_validator(common_config, False)
    except GreaterThanSupportedQualibrateConfigVersionError as ex:
        error_msg = (
            f"QUAlibrate was unable to load the config. {str(ex)}. If this "
            f'problem persists, please delete "{config_path}" and retry '
            'running "qualibrate config"'
        )
        raise RuntimeError(error_msg) from ex
    except InvalidQualibrateConfigVersionError:
        if common_config:
            run_migrations(config_path)
            return get_config_file_content(config_path)
    return common_config, config_path


def get_config_solved_references_or_print_error(
    config_path: Path,
) -> RawConfigType | None:
    try:
        return read_config_file(config_path, solve_references=True)
    except ValueError as ex:
        click.secho(str(ex), fg="red")
        click.secho(SUGGEST_MSG, fg="yellow")
    return None


def get_config_model_or_print_error(
    config: RawConfigType,
    model_type: type[T],
    config_key: str | None,
) -> T | None:
    try:
        return model_type(config)
    except ValidationError as ex:
        prefix = [config_key] if config_key else []
        errors = [
            (
                f"Message: {error.get('msg')}. "
                "Path: "
                f"{'.'.join([*prefix, *map(str, error.get('loc', []))])}. "
                f"Value: {error.get('input')}"
            )
            for error in ex.errors()
        ]
        click.secho("\n".join(errors), fg="red")
        click.secho(SUGGEST_MSG, fg="yellow")
    return None



import shutil
from collections.abc import Iterable, Mapping
from copy import deepcopy
from pathlib import Path
from typing import Any

import jsonpatch
import tomli_w
from click import Context

from qualibrate_config.core.content import get_config_file_content
from qualibrate_config.core.from_sources import qualibrate_config_from_sources
from qualibrate_config.core.project.p_list import list_projects
from qualibrate_config.core.project.path import (
    get_project_config_path,
    get_project_path,
)
from qualibrate_config.models import PathSerializer, QualibrateTopLevelConfig
from qualibrate_config.validation import validate_version_and_migrate_if_needed
from qualibrate_config.vars import (
    QUALIBRATE_CONFIG_KEY,
    QUAM_CONFIG_KEY,
    QUAM_STATE_PATH_CONFIG_KEY,
)


def jsonpatch_to_dict(patch: jsonpatch.JsonPatch) -> dict[str, Any]:
    d: dict[str, Any] = dict()
    replace_ops = filter(
        lambda op: isinstance(op, jsonpatch.ReplaceOperation), patch._ops
    )
    for op in replace_ops:
        path_ = d
        for part in op.pointer.parts[:-1]:
            if part not in path_:
                path_[part] = {}
            path_ = path_[part]
        path_[op.key] = op.operation["value"]
    return d


def _fill_path(
    config: dict[str, Any], path: Iterable[str], key: str, value: Path | None
) -> None:
    if value is None:
        return
    c_item = config
    for p in path:
        c_item = c_item.setdefault(p, {})
    c_item[key] = PathSerializer.serialize_path(value)


def fill_project_quam_state_path(
    common_config: dict[str, Any], quam_state_path: Path | None
) -> None:
    _fill_path(
        common_config,
        [QUAM_CONFIG_KEY],
        QUAM_STATE_PATH_CONFIG_KEY,
        quam_state_path,
    )


def fill_project_storage_location(
    common_config: dict[str, Any], storage_location: Path | None
) -> None:
    _fill_path(
        common_config,
        [QUALIBRATE_CONFIG_KEY, "storage"],
        "location",
        storage_location,
    )


def fill_project_calibration_library_folder(
    common_config: dict[str, Any], calibration_library_folder: Path | None
) -> None:
    _fill_path(
        common_config,
        [QUALIBRATE_CONFIG_KEY, "calibration_library"],
        "folder",
        calibration_library_folder,
    )


def after_create_project(
    storage_location: Path | None,
    quam_state_path: Path | None,
) -> None:
    if storage_location:
        storage_location.mkdir(parents=True, exist_ok=True)
    if quam_state_path:
        quam_state_path.mkdir(parents=True, exist_ok=True)


def create_project_config_file(
    qualibrate_path: Path,
    project_name: str,
    config_overrides: Mapping[str, Any] | None = None,
) -> None:
    config_filepath = get_project_config_path(qualibrate_path, project_name)
    config_filepath.parent.mkdir(parents=True, exist_ok=True)
    if config_overrides:
        with config_filepath.open("wb") as f_out:
            tomli_w.dump(config_overrides, f_out)
    else:
        config_filepath.touch()


def rollback_project_creation(
    qualibrate_path: Path,
    project_name: str,
    storage_path: Path | None,
    quam_state_path: Path | None,
) -> None:
    project_path = get_project_path(qualibrate_path, project_name)
    for p in (project_path, storage_path, quam_state_path):
        if p and p.is_dir():
            shutil.rmtree(p)


def config_for_project_from_context(
    common_config: dict[str, Any],
    storage_location: Path | None,
    calibration_library_folder: Path | None,
    quam_state_path: Path | None,
    context: Context | None,
) -> dict[str, Any]:
    if context is None:
        raise ValueError("Context isn't passed.")
    q_config = common_config.get(QUALIBRATE_CONFIG_KEY, {})
    required_subconfigs = ("storage",)
    optional_subconfigs = (
        ("calibration_library",)
        if (
            calibration_library_folder
            or q_config.get("calibration_library") is not None
        )
        else tuple()
    )
    q_config = qualibrate_config_from_sources(
        context,
        q_config,
        required_subconfigs,
        optional_subconfigs,
    )
    if (
        calibration_library_folder is not None
        and q_config.get("calibration_library", {}).get("resolver") is None
    ):
        raise ValueError(
            "Calibration library folder can't be specified without a resolver."
        )
    qs = QualibrateTopLevelConfig({QUALIBRATE_CONFIG_KEY: q_config})
    common_config.update(qs.serialize())
    fill_project_quam_state_path(common_config, quam_state_path)
    return common_config


def config_for_project_from_args(
    common_config: dict[str, Any],
    storage_location: Path | None,
    calibration_library_folder: Path | None,
    quam_state_path: Path | None,
    context: Context | None,
) -> dict[str, Any]:
    if context is not None:
        raise ValueError("Context is passed.")
    fill_project_storage_location(common_config, storage_location)
    fill_project_calibration_library_folder(
        common_config, calibration_library_folder
    )
    fill_project_quam_state_path(common_config, quam_state_path)
    return common_config


def create_project(
    config_path: Path,
    name: str,
    storage_location: Path | None,
    calibration_library_folder: Path | None,
    quam_state_path: Path | None,
    ctx: Context | None = None,
) -> None:
    qualibrate_path = config_path.parent
    try:
        projects = list_projects(qualibrate_path)
    except NotADirectoryError:
        pass
    else:
        if name in projects:
            raise ValueError(f"Project '{name}' already exists.")
    common_config, config_file = get_config_file_content(config_path)
    common_config, config_file = validate_version_and_migrate_if_needed(
        common_config, config_file
    )
    old_config = deepcopy(common_config)
    config_for_project = (
        config_for_project_from_context if ctx else config_for_project_from_args
    )
    common_config = config_for_project(
        common_config,
        storage_location,
        calibration_library_folder,
        quam_state_path,
        ctx,
    )
    patches = jsonpatch.make_patch(old_config, common_config)
    project_config = jsonpatch_to_dict(patches)
    try:
        create_project_config_file(qualibrate_path, name, project_config)
        after_create_project(storage_location, quam_state_path)
    except Exception as exc:
        rollback_project_creation(
            qualibrate_path, name, storage_location, quam_state_path
        )
        raise ValueError(f"Project creation failed. {exc}") from exc



from copy import deepcopy
from pathlib import Path

from qualibrate_config.core.migration.migrations.base import MigrateBase
from qualibrate_config.qulibrate_types import RawConfigType


class Migrate(MigrateBase):
    from_version: int = 1
    to_version: int = 2

    @staticmethod
    def backward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        qualibrate = data.pop("qualibrate")
        quam: RawConfigType | None = data.pop("quam", None)
        assert qualibrate.pop("version") == Migrate.to_version
        qualibrate["config_version"] = 1

        app = qualibrate.pop("app", None)
        runner = qualibrate.pop("runner", None)
        composite = qualibrate.pop("composite", None)
        calibration_lib = qualibrate.pop("calibration_library", None)
        new_data = {"qualibrate": qualibrate}
        if app:
            if runner:
                app["runner"] = runner
            new_data["qualibrate_app"] = app
        if composite:
            if runner:
                composite["runner"].update(runner)
            new_data["qualibrate_composite"] = composite
        if calibration_lib:
            new_data["qualibrate_runner"] = {
                "calibration_library_resolver": calibration_lib["resolver"],
                "calibration_library_folder": calibration_lib["folder"],
            }
        if quam is not None and "state_path" in quam:
            new_data["active_machine"] = {"path": quam["state_path"]}
        return new_data

    @staticmethod
    def forward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        new_qualibrate = data.pop("qualibrate")
        new_qualibrate.pop("config_version")
        q_app: RawConfigType | None = data.pop("qualibrate_app", None)
        q_composite: RawConfigType | None = data.pop(
            "qualibrate_composite", None
        )
        q_runner: RawConfigType | None = data.pop("qualibrate_runner", None)
        active_machine: RawConfigType | None = data.get("active_machine")
        if active_machine is not None and "path" in active_machine:
            data["quam"] = {"state_path": active_machine["path"]}

        if q_app:
            new_qualibrate["app"] = {
                "static_site_files": q_app["static_site_files"]
            }
        if remote_runner := (
            (q_composite and q_composite.get("runner"))
            or (q_app and q_app.get("runner"))
        ):
            new_runner = {}
            if "address" in remote_runner:
                new_runner["address"] = remote_runner["address"]
            if "timeout" in remote_runner:
                new_runner["timeout"] = remote_runner["timeout"]
            if new_runner:
                new_qualibrate["runner"] = new_runner
        if q_composite:
            new_qualibrate["composite"] = {
                "app": {"spawn": q_composite["app"]["spawn"]},
                "runner": {"spawn": q_composite["runner"]["spawn"]},
            }
        if q_runner:
            new_qualibrate["calibration_library"] = {
                "resolver": q_runner["calibration_library_resolver"],
                "folder": q_runner["calibration_library_folder"],
            }
        new_qualibrate["version"] = Migrate.to_version
        new_data = {"qualibrate": new_qualibrate, **data}
        return new_data



import sys
from pathlib import Path

import click

from qualibrate_config.cli.vars import (
    CALIBRATION_LIBRARY_FOLDER_HELP,
    CONFIG_PATH_HELP,
    QUAM_STATE_PATH_HELP,
    STORAGE_LOCATION_HELP,
)
from qualibrate_config.core.project.create import (
    create_project,
)
from qualibrate_config.vars import (
    DEFAULT_CONFIG_FILEPATH,
)

__all__ = ["create_command"]


@click.command(name="create")
@click.argument("name", type=str)
@click.option(
    "--config-path",
    type=click.Path(exists=True, path_type=Path),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
@click.option(
    "--storage-location",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    required=False,
    help=STORAGE_LOCATION_HELP,
)
@click.option(
    "--calibration-library-folder",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    required=False,
    help=CALIBRATION_LIBRARY_FOLDER_HELP,
)
@click.option(
    "--quam-state-path",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    required=False,
    help=QUAM_STATE_PATH_HELP,
)
@click.pass_context
def create_command(
    ctx: click.Context,
    name: str,
    config_path: Path,
    storage_location: Path | None,
    calibration_library_folder: Path | None,
    quam_state_path: Path | None,
) -> None:
    try:
        create_project(
            config_path,
            name,
            storage_location,
            calibration_library_folder,
            quam_state_path,
            ctx,
        )
    except ValueError as e:
        click.secho(str(e), fg="red")
        sys.exit(1)


if __name__ == "__main__":
    create_command(["test_creation"], standalone_mode=False)



import sys
from pathlib import Path

import click

from qualibrate_config.cli.vars import CONFIG_PATH_HELP
from qualibrate_config.core.project.active import get_active_project
from qualibrate_config.vars import DEFAULT_CONFIG_FILEPATH

__all__ = ["current_command"]


@click.command(name="current")
@click.option(
    "--config-path",
    type=click.Path(exists=True, path_type=Path),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
def current_command(
    config_path: Path,
) -> None:
    project = get_active_project(config_path)
    if project is None:
        click.secho(
            "Can't resolve currently active project. If you have no "
            "project please create it. Otherwise please regenerate config "
            "using `qualibrate-config config` command.",
            fg="yellow",
        )
        sys.exit(1)
    click.echo(f"Current project is {project}")


if __name__ == "__main__":
    current_command([], standalone_mode=False)



import os
from collections.abc import Mapping
from itertools import filterfalse
from pathlib import Path
from typing import IO, Any

import click
from click.exceptions import Exit

from qualibrate_config.qulibrate_types import RawConfigType


def print_config(
    data: Mapping[str, Any], depth: int = 0, file: IO[Any] | None = None
) -> None:
    if not len(data.keys()):
        return
    max_key_len = max(map(len, map(str, data.keys())))
    non_mapping_items = list(
        filterfalse(lambda item: isinstance(item[1], Mapping), data.items())
    )
    if len(non_mapping_items):
        click.echo(
            os.linesep.join(
                f"{' ' * 4 * depth}{f'{k} :':<{max_key_len + 3}} {v}"
                for k, v in non_mapping_items
            ),
            file=file,
        )
    mappings = filter(lambda x: isinstance(x[1], Mapping), data.items())
    for mapping_k, mapping_v in mappings:
        click.echo(f"{' ' * 4 * depth}{mapping_k} :", file=file)
        print_config(mapping_v, depth + 1, file=file)


def print_and_confirm(
    config_file: Path,
    exported_data: RawConfigType,
    check_generator: bool,
) -> None:
    click.echo(f"Config file path: {config_file}")
    click.echo(click.style("Generated config:", bold=True))
    print_config(exported_data)
    if check_generator:
        raise Exit(0)
    confirmed = click.confirm("Do you confirm config?", default=True)
    if not confirmed:
        click.echo(
            click.style(
                (
                    "The configuration has not been confirmed. "
                    "Rerun config script."
                ),
                fg="yellow",
            )
        )
        raise Exit(1)



from copy import deepcopy
from importlib.util import find_spec
from pathlib import Path

from qualibrate_config.core.migration.migrations.base import MigrateBase
from qualibrate_config.qulibrate_types import RawConfigType


class Migrate(MigrateBase):
    from_version: int = 3
    to_version: int = 4

    @staticmethod
    def backward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        qualibrate = data.pop("qualibrate")
        assert qualibrate.pop("version") == Migrate.to_version
        qualibrate["version"] = Migrate.from_version
        app: RawConfigType | None = qualibrate.get("app", None)
        if app is None:
            return {"qualibrate": qualibrate, **data}
        static_files = app.get("static_site_files")
        if static_files is not None:
            return {"qualibrate": qualibrate, **data}
        app_module = find_spec("qualibrate_app")
        if app_module is None or app_module.origin is None:
            return {"qualibrate": qualibrate, **data}
        module_path = Path(app_module.origin)
        app["static_site_files"] = str(
            module_path.parents[1] / "qualibrate_static"
        )
        return {"qualibrate": qualibrate, **data}

    @staticmethod
    def forward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        new_qualibrate = data.pop("qualibrate")
        version = new_qualibrate.pop("version")
        assert version == Migrate.from_version
        new_qualibrate["version"] = Migrate.to_version
        app: RawConfigType | None = new_qualibrate.get("app", None)

        if app is None:
            return {"qualibrate": new_qualibrate, **data}
        static_files = app.get("static_site_files")
        if static_files is None:
            return {"qualibrate": new_qualibrate, **data}
        path = Path(static_files)
        if tuple(path.parts[-2:]) == ("site-packages", "qualibrate_static"):
            app.pop("static_site_files", None)
        return {"qualibrate": new_qualibrate, **data}



from pathlib import Path

import click

from qualibrate_config.cli.vars import CONFIG_PATH_HELP
from qualibrate_config.core.migration.migrate import run_migrations
from qualibrate_config.models import QualibrateConfig
from qualibrate_config.vars import DEFAULT_CONFIG_FILEPATH

__all__ = ["migrate_command"]


@click.command(name="migrate", help="Migrate qualibrate config files")
@click.option(
    "--config-path",
    type=click.Path(
        exists=False,
        path_type=Path,
    ),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
@click.argument("to_version", type=int, default=QualibrateConfig.version)
def migrate_command(config_path: Path, to_version: int) -> None:
    run_migrations(
        config_path=config_path,
        to_version=to_version,
    )


if __name__ == "__main__":
    migrate_command([], standalone_mode=False)



from typing import Any

from pydantic import BaseModel, Field


class Reference(BaseModel):
    config_path: str
    reference_path: str
    index_start: int
    index_end: int
    value: Any = None
    solved: bool = False

    def __hash__(self) -> int:
        return hash(
            (
                self.config_path,
                self.reference_path,
                self.index_start,
                self.index_end,
            )
        )


class PathWithSolvingReferences(BaseModel):
    config_path: str
    value: Any = None
    solved: bool = False
    references: list[Reference] = Field(default_factory=list)



from copy import deepcopy
from pathlib import Path

import click
from click.exceptions import Exit

from qualibrate_config.cli.deprecated import (
    DeprecatedOption,
    DeprecatedOptionsCommand,
)
from qualibrate_config.cli.vars import (
    CALIBRATION_LIBRARY_FOLDER_HELP,
    CONFIG_PATH_HELP,
    QUAM_STATE_PATH_HELP,
    STORAGE_LOCATION_HELP,
)
from qualibrate_config.core.content import (
    get_config_file_content,
    simple_write,
    write_config,
)
from qualibrate_config.core.defaults import get_user_storage
from qualibrate_config.core.from_sources import (
    qualibrate_config_from_sources,
)
from qualibrate_config.models import PathSerializer
from qualibrate_config.models.qualibrate import QualibrateTopLevelConfig
from qualibrate_config.models.storage_type import StorageType
from qualibrate_config.qulibrate_types import RawConfigType
from qualibrate_config.validation import (
    validate_version_and_migrate_if_needed,
)
from qualibrate_config.vars import (
    DEFAULT_CONFIG_FILEPATH,
    QUALIBRATE_CONFIG_KEY,
    QUALIBRATE_PATH,
    QUAM_CONFIG_KEY,
    QUAM_STATE_PATH_CONFIG_KEY,
)

__all__ = ["config_command"]


@click.command(
    name="config",
    cls=DeprecatedOptionsCommand,
    help="Common config definition",
)
@click.option(
    "--config-path",
    type=click.Path(
        exists=False,
        path_type=Path,
    ),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
@click.option(
    "--auto-accept",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "Flag responsible for whether to skip confirmation of the generated "
        "config."
    ),
)
@click.option(
    "--project",
    type=str,
    default="init_project",
    show_default=True,
    help=(
        "The name of qualibrate app project that will be used for storing runs "
        "results and resolving them."
    ),
)
@click.option(
    "--log-folder",
    type=click.Path(file_okay=False, resolve_path=True, path_type=Path),
    default=QUALIBRATE_PATH / "logs",
    help="The path to the directory where the logs should be stored to.",
    show_default=True,
)
@click.option(
    "--password",
    "--qualibrate-password",
    type=str,
    default=None,
    cls=DeprecatedOption,
    deprecated=("--qualibrate-password",),
    preferred="--password",
    help=(
        "Password used to authorize users. By default, no password is used. "
        "Everyone has access to the API. If a password is specified during "
        "configuration, you must log in to access the API."
    ),
)
@click.option(
    "--storage-type",
    type=click.Choice([t.value for t in StorageType]),
    default="local_storage",
    show_default=True,
    callback=lambda ctx, param, value: StorageType(value),
    help=(
        "Type of storage. Only `local_storage` is supported now. Use specified "
        "local path as the database."
    ),
)
@click.option(
    "--storage-location",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    default=get_user_storage(),
    show_default=True,
    help=STORAGE_LOCATION_HELP,
)
@click.option(
    "--calibration-library-resolver",
    "--runner-calibration-library-resolver",
    type=str,
    default="qualibrate.QualibrationLibrary",
    show_default=True,
    cls=DeprecatedOption,
    deprecated=("--runner-calibration-library-resolver",),
    preferred="--calibration-library-resolver",
    help=(
        "String contains python path to the class that represents qualibration "
        "library."
    ),
)
@click.option(
    "--calibration-library-folder",
    "--runner-calibration-library-folder",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    default=QUALIBRATE_PATH / "calibrations",
    show_default=True,
    cls=DeprecatedOption,
    deprecated=("--runner-calibration-library-folder",),
    preferred="--calibration-library-folder",
    help=CALIBRATION_LIBRARY_FOLDER_HELP,
)
@click.option(
    "--spawn-runner",
    type=bool,
    default=True,
    show_default=True,
    help=(
        "This flag indicates whether the `qualibrate-runner` service should be "
        "started. This service is designed to run nodes and graphs. The service"
        " can be spawned independently."
    ),
)
@click.option(
    "--runner-address",
    type=str,  # TODO: add type check for addr
    default="http://127.0.0.1:8001/execution/",
    show_default=True,
    help=(
        "Address of `qualibrate-runner` service. If the service is spawned by "
        "the `qualibrate` then the default address should be kept as is. If you"
        " are running the service separately, you must specify its address."
    ),
)
@click.option(
    "--runner-timeout",
    type=float,
    default=1.0,
    show_default=True,
    help=(
        "Maximum waiting time for a response from the `qualibrate-runner` "
        "service."
    ),
)
@click.option(
    "--spawn-app",
    type=bool,
    default=True,
    show_default=True,
    help=(
        "This flag indicates whether the `qualibrate-app` service should be "
        "started. This service is designed to getting info about snapshots. "
        "The service can be spawned independently."
    ),
)
@click.option(
    "--app-static-site-files",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    help="Path to the frontend build static files.",
)
@click.option(
    "--spawn-qua-dashboards",
    type=bool,
    default=True,
    show_default=True,
    help=(
        "This flag indicates whether the `qua-dashboards` service should be "
        "started."
    ),
)
@click.option(
    "--quam-state-path",
    "--active-machine-path",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    cls=DeprecatedOption,
    required=False,
    deprecated=("--active-machine-path",),
    preferred="--quam-state-path",
    help=QUAM_STATE_PATH_HELP,
)
@click.option("--check-generator", is_flag=True, hidden=True)
@click.pass_context
def config_command(
    ctx: click.Context,
    config_path: Path,
    auto_accept: bool,
    project: str,
    log_folder: Path,
    password: str | None,
    storage_type: StorageType,
    storage_location: Path,
    calibration_library_resolver: str,
    calibration_library_folder: Path,
    spawn_runner: bool,
    runner_address: str,
    runner_timeout: float,
    spawn_app: bool,
    app_static_site_files: Path,
    spawn_qua_dashboards: bool,
    quam_state_path: Path | None,
    check_generator: bool,
) -> None:
    common_config, config_file = get_config_file_content(config_path)
    old_config = deepcopy(common_config)
    common_config, config_file = validate_version_and_migrate_if_needed(
        common_config, config_file
    )
    qualibrate_config = common_config.get(QUALIBRATE_CONFIG_KEY, {})
    required_subconfigs = ("storage",)
    optional_subconfigs = ("app", "runner", "composite", "calibration_library")
    qualibrate_config = qualibrate_config_from_sources(
        ctx,
        qualibrate_config,
        required_subconfigs,
        optional_subconfigs,
    )
    qs = QualibrateTopLevelConfig({QUALIBRATE_CONFIG_KEY: qualibrate_config})
    qs.qualibrate.storage.location.mkdir(parents=True, exist_ok=True)
    try:
        _temporary_fill_quam_state_path(common_config, quam_state_path)
        write_config(
            config_file,
            common_config,
            qs.qualibrate,
            QUALIBRATE_CONFIG_KEY,
            confirm=not auto_accept,
            check_generator=check_generator,
        )
    except Exit:
        if old_config:
            simple_write(config_file, old_config)
        raise


def _temporary_fill_quam_state_path(
    common_config: RawConfigType, state_path: Path | None
) -> RawConfigType:
    quam = common_config.setdefault(QUAM_CONFIG_KEY, {})
    active_machine = common_config.get("active_machine", {})
    new_state_path = (
        state_path
        or quam.get(QUAM_STATE_PATH_CONFIG_KEY)
        or active_machine.get("path")
    )
    if new_state_path is None:
        return common_config
    quam[QUAM_STATE_PATH_CONFIG_KEY] = PathSerializer.serialize_path(
        new_state_path
    )
    return common_config


if __name__ == "__main__":
    config_command([], standalone_mode=False)



import shutil
from copy import deepcopy
from pathlib import Path
from urllib.parse import urlparse, urlunparse

from qualibrate_config.core.migration.migrations.base import MigrateBase
from qualibrate_config.core.project.create import create_project_config_file
from qualibrate_config.core.project.path import get_projects_path
from qualibrate_config.qulibrate_types import RawConfigType


class Migrate(MigrateBase):
    from_version: int = 4
    to_version: int = 5

    @staticmethod
    def backward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        qualibrate = data.pop("qualibrate")
        assert qualibrate.pop("version") == Migrate.to_version
        qualibrate["version"] = Migrate.from_version
        shutil.rmtree(get_projects_path(config_path.parent), ignore_errors=True)
        return {"qualibrate": qualibrate, **data}

    @staticmethod
    def forward(data: RawConfigType, config_path: Path) -> RawConfigType:
        data = deepcopy(data)
        qualibrate = data.pop("qualibrate")
        assert qualibrate.pop("version") == Migrate.from_version
        qualibrate["version"] = Migrate.to_version
        if (runner := qualibrate.get("runner")) and (
            address := runner.get("address")
        ):
            parsed = urlparse(address)
            if parsed.hostname == "localhost":
                runner["address"] = urlunparse(
                    parsed._replace(
                        netloc=parsed.netloc.replace("localhost", "127.0.0.1")
                    )
                )
        project_name = qualibrate.get("project")
        new_config = {"qualibrate": qualibrate, **data}
        if project_name is None:
            return new_config
        create_project_config_file(config_path.parent, project_name)
        return new_config



from .base import BaseConfig, Importable, PathSerializer
from .calibration_library import CalibrationLibraryConfig
from .composite import QualibrateCompositeConfig
from .q_app import QualibrateAppConfig
from .qualibrate import QualibrateConfig, QualibrateTopLevelConfig
from .remote_services import (
    JsonTimelineDBRemoteServiceConfig,
    QuaDashboardSubServiceConfig,
    QualibrateAppSubServiceConfig,
    QualibrateRunnerRemoteServiceConfig,
    QualibrateRunnerSubServiceConfig,
    RemoteServiceBaseConfig,
    SpawnServiceBaseConfig,
)
from .storage import StorageConfig
from .storage_type import StorageType

__all__ = [
    "PathSerializer",
    "Importable",
    "BaseConfig",
    "CalibrationLibraryConfig",
    "QualibrateCompositeConfig",
    "QualibrateAppConfig",
    "QualibrateConfig",
    "QualibrateTopLevelConfig",
    "JsonTimelineDBRemoteServiceConfig",
    "RemoteServiceBaseConfig",
    "SpawnServiceBaseConfig",
    "QuaDashboardSubServiceConfig",
    "QualibrateAppSubServiceConfig",
    "QualibrateRunnerSubServiceConfig",
    "QualibrateRunnerRemoteServiceConfig",
    "StorageConfig",
    "StorageType",
]



from pathlib import Path

from qualibrate_config.vars import DEFAULT_CONFIG_FILENAME


def get_projects_path(qualibrate_path: Path) -> Path:
    return qualibrate_path / "projects"


def get_project_path(qualibrate_path: Path, project_name: str) -> Path:
    return get_projects_path(qualibrate_path) / project_name


def get_project_config_path(
    qualibrate_path: Path,
    project_name: str,
    config_filename: str = DEFAULT_CONFIG_FILENAME,
) -> Path:
    return get_project_path(qualibrate_path, project_name).joinpath(
        config_filename
    )



import click

from qualibrate_config.cli import (
    config_command,
    migrate_command,
    project_group,
)


@click.group()
def cli() -> None:
    pass


cli.add_command(config_command)
cli.add_command(migrate_command)
cli.add_command(project_group)


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



import sys
from pathlib import Path

import click

__all__ = ["delete_command"]

from qualibrate_config.cli.vars import CONFIG_PATH_HELP
from qualibrate_config.core.project.delete import delete_project
from qualibrate_config.vars import DEFAULT_CONFIG_FILEPATH


@click.command(name="delete")
@click.argument("name", type=str)
@click.option(
    "--config-path",
    type=click.Path(exists=True, path_type=Path),
    default=DEFAULT_CONFIG_FILEPATH,
    show_default=True,
    help=CONFIG_PATH_HELP,
)
def delete_command(
    name: str,
    config_path: Path,
) -> None:
    try:
        delete_project(config_path, name)
    except RuntimeError as e:
        click.secho(f"Failed to remove project '{name}'. {e}", fg="red")
        sys.exit(1)
    else:
        click.echo(f"Successfully removed project '{name}'")


if __name__ == "__main__":
    delete_command([], standalone_mode=False)



import datetime
import sys
from os import stat_result
from pathlib import Path

import click

from qualibrate_config.core.project.common import read_project_config_file
from qualibrate_config.core.project.model import Project
from qualibrate_config.core.project.path import (
    get_project_path,
    get_projects_path,
)
from qualibrate_config.file import read_config_file
from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY


def _dt_from_ts(ts: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(round(ts))


def _stat_ctime(storage_stat: stat_result) -> float:
    if sys.version_info >= (3, 12) and sys.platform == "win32":
        return storage_stat.st_birthtime
    return storage_stat.st_ctime


def _file_stat_ctime(p: Path) -> float:
    return _stat_ctime(p.stat())


def _file_stat_mtime(p: Path) -> float:
    return p.stat().st_mtime


def _created_at_from_storage(
    storage_path: Path, storage_stat: stat_result
) -> datetime.datetime:
    first_create_p = min(
        (p for p in storage_path.glob("*/#*") if p.is_dir()),
        key=_file_stat_ctime,
        default=None,
    )
    if first_create_p is None:
        return _dt_from_ts(_stat_ctime(storage_stat))
    return _dt_from_ts(
        min(_stat_ctime(storage_stat), _file_stat_ctime(first_create_p))
    )


def _last_modified_at_from_storage(
    storage_path: Path, storage_stat: stat_result
) -> datetime.datetime:
    first_create_p = max(
        (p for p in storage_path.glob("*/#*") if p.is_dir()),
        key=_file_stat_mtime,
        default=None,
    )
    if first_create_p is None:
        return _dt_from_ts(storage_stat.st_mtime)
    return _dt_from_ts(
        max(storage_stat.st_mtime, _file_stat_mtime(first_create_p))
    )


def _storage_stat(
    storage_path: Path,
) -> tuple[int, datetime.datetime, datetime.datetime]:
    nodes_number = sum(
        (1 for p in storage_path.glob("*/#*") if p.is_dir()),
        start=0,
    )
    storage_stat = storage_path.stat()
    created_at = _created_at_from_storage(storage_path, storage_stat)
    last_modified_at = _last_modified_at_from_storage(
        storage_path, storage_stat
    )
    return nodes_number, created_at, last_modified_at


def _project_stat_dir(
    project_path: Path,
) -> tuple[int, datetime.datetime, datetime.datetime]:
    stat = project_path.stat()
    return 0, _dt_from_ts(stat.st_ctime), _dt_from_ts(stat.st_mtime)


def project_stat(
    qualibrate_path: Path, project: str, config_path: Path
) -> Project:
    project_path = get_project_path(qualibrate_path, project)

    config_dict = read_config_file(config_path, override_project=project)
    project_config = read_project_config_file(config_path, project)
    storage_location = (
        config_dict.get(QUALIBRATE_CONFIG_KEY, {})
        .get("storage", {})
        .get("location")
    )
    if storage_location:
        storage_path = Path(storage_location)
        if storage_path.is_dir():
            nodes_number, created_at, last_modified_at = _storage_stat(
                storage_path
            )
        else:
            nodes_number, created_at, last_modified_at = _project_stat_dir(
                project_path
            )
    else:
        nodes_number, created_at, last_modified_at = _project_stat_dir(
            project_path
        )
    return Project(
        name=project,
        nodes_number=nodes_number,
        created_at=created_at.astimezone(),
        last_modified_at=last_modified_at.astimezone(),
        updates=project_config,
    )


def list_projects(qualibrate_path: Path) -> list[str]:
    projects_path = get_projects_path(qualibrate_path)
    if not projects_path.is_dir():
        raise NotADirectoryError(
            f"Projects path '{projects_path}' is not a directory"
        )
    return list(
        map(
            lambda p: p.name,
            filter(Path.is_dir, projects_path.iterdir()),
        )
    )


def verbose_list_projects(
    config_path: Path,
) -> dict[str, Project]:
    qualibrate_path = config_path.parent
    return {
        p_name: project_stat(qualibrate_path, p_name, config_path)
        for p_name in list_projects(qualibrate_path)
    }


def print_simple_projects_list(
    config_path: Path,
) -> None:
    click.echo("\n".join(list_projects(config_path.parent)))


def print_verbose_projects_list(
    config_path: Path,
) -> None:
    for p_stat in verbose_list_projects(config_path).values():
        click.echo(p_stat.verbose_str())



import functools
import warnings

from qualibrate_config.cli.utils import deprecated_alias
from qualibrate_config.core.migration import utils as _migration_utils

__all__ = _migration_utils.__all__

deprecated_m = "qualibrate_config.cli.utils.migration_utils"
new_m = "qualibrate_config.core.migration.utils"
warnings.warn(
    (
        f"Module '{deprecated_m}' is deprecated and will be removed in a "
        f"future version. Please use '{new_m}' instead."
    ),
    DeprecationWarning,
    stacklevel=2,
)

l_deprecated = functools.partial(
    deprecated_alias, deprecated_module=deprecated_m, new_module=new_m
)

# Apply decorator to each re-exported function
make_migrations = l_deprecated(
    name="make_migrations",
)(_migration_utils.make_migrations)



from abc import ABC, abstractmethod
from pathlib import Path

from qualibrate_config.qulibrate_types import RawConfigType


class MigrateBase(ABC):
    from_version: int
    to_version: int

    @staticmethod
    @abstractmethod
    def backward(data: RawConfigType, config_path: Path) -> RawConfigType:
        pass

    @staticmethod
    @abstractmethod
    def forward(data: RawConfigType, config_path: Path) -> RawConfigType:
        pass



from typing import Any

import click

__all__ = ["DeprecatedOption"]


class DeprecatedOption(click.Option):
    def __init__(
        self,
        *args: Any,
        preferred: str | None = None,
        deprecated: tuple[str] | None = None,
        **kwargs: Any,
    ):
        self.__deprecated = deprecated or ()
        self.__preferred = preferred or args[0][-1]
        super().__init__(*args, **kwargs)

    @property
    def deprecated_preferred(self) -> tuple[tuple[str, ...], str]:
        return self.__deprecated, self.__preferred



from pathlib import Path

from qualibrate_config.models.base.config_base import BaseConfig
from qualibrate_config.models.base.importable import Importable

__all__ = ["CalibrationLibraryConfig"]


class CalibrationLibraryConfig(BaseConfig):
    folder: Path

    resolver: Importable



from collections.abc import Callable, Iterable, Mapping
from typing import Any, Protocol, TypeVar, overload

from qualibrate_config.qulibrate_types import RawConfigType


def recursive_update_dict(
    to_update: RawConfigType,
    updates: Mapping[str, Any],
) -> RawConfigType:
    for k, v in updates.items():
        if k not in to_update:
            continue
        if isinstance(v, Mapping):
            to_update[k] = recursive_update_dict(to_update.get(k, {}), v)
        else:
            to_update[k] = v
    return to_update


class SupportsRichComparison(Protocol):
    def __lt__(self, other: Any) -> bool: ...

    def __gt__(self, other: Any) -> bool: ...


T = TypeVar("T")
KT = TypeVar("KT", bound=SupportsRichComparison)
CT = TypeVar("CT", bound=SupportsRichComparison)
R1 = TypeVar("R1")
R2 = TypeVar("R2")


@overload
def minmax(
    iterable: Iterable[CT],
    *,
    key: None = ...,
    default: tuple[R1, R2] | None = ...,
) -> tuple[CT | R1, CT | R2]: ...


@overload
def minmax(
    iterable: Iterable[T],
    *,
    key: Callable[[T], KT],
    default: tuple[R1, R2] | None = ...,
) -> tuple[T | R1, T | R2]: ...


def minmax(
    iterable: Iterable[Any],
    *,
    key: Callable[[Any], Any] | None = None,
    default: tuple[Any, Any] | None = None,
) -> tuple[Any, Any]:
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        if default is not None:
            return default
        raise ValueError("minmax() arg is an empty sequence") from None

    if key is None:
        min_key = max_key = first
        for elem in it:
            min_key, max_key, _, _ = _minmax(
                min_key, max_key, elem, min_key, max_key, elem
            )
        return min_key, max_key
    else:
        min_elem = max_elem = first
        min_key = max_key = key(first)
        for elem in it:
            min_key, max_key, min_elem, max_elem = _minmax(
                min_key, max_key, key(elem), min_elem, max_elem, elem
            )
        return min_elem, max_elem


def _minmax(
    min_key: KT,
    max_key: KT,
    k: KT,
    min_elem: T,
    max_elem: T,
    elem: T,
) -> tuple[KT, KT, T, T]:
    if k < min_key:
        min_elem, min_key = elem, k
    elif k > max_key:
        max_elem, max_key = elem, k
    return min_key, max_key, min_elem, max_elem



from collections.abc import Callable, Sequence
from types import MethodType
from typing import Any

import click
from click.core import Option as CoreOption
from click.parser import Option as ParserOption
from click.parser import ParsingState

from qualibrate_config.cli.deprecated.deprecated_option import DeprecatedOption

__all__ = ["DeprecatedOptionsCommand"]


def _validate_deprecated_option(
    option: CoreOption,
) -> tuple[Sequence[str], str]:
    assert isinstance(option, DeprecatedOption)
    deprecated, preferred = option.deprecated_preferred
    msg = "Expected `deprecated` value for `{}`"
    assert deprecated is not None, msg.format(option.name)
    return deprecated, preferred


def make_deprecated_process(
    option: ParserOption,
) -> Callable[[ParserOption, Any, ParsingState], None]:
    """Construct a closure to the parser option processor"""
    option_instance = option.obj
    original_process = option.process
    deprecated, preferred = _validate_deprecated_option(option_instance)

    def process(self: ParserOption, value: Any, state: ParsingState) -> None:
        """The function above us on the stack used 'opt' to
        pick option from a dict, see if it is deprecated"""
        # reach up the stack and get 'opt'
        import inspect

        frame = inspect.currentframe()
        if frame is None:
            return original_process(value, state)
        opt = None
        try:
            if frame.f_back is not None:
                opt = frame.f_back.f_locals.get("opt")
        finally:
            del frame
        if opt in deprecated:
            msg = "'{}' has been deprecated, use '{}'"
            click.secho(msg.format(opt, preferred), fg="yellow")

        return original_process(value, state)

    return process


class DeprecatedOptionsCommand(click.Command):
    def make_parser(self, ctx: click.Context) -> click.OptionParser:
        """Hook 'make_parser' and during processing check the name
        used to invoke the option to see if it is preferred"""

        parser = super().make_parser(ctx)

        # get the parser options
        options = set(parser._short_opt.values())
        options |= set(parser._long_opt.values())
        for option in options:
            if isinstance(option.obj, DeprecatedOption):
                option.process = MethodType(  # type: ignore
                    make_deprecated_process(option), option
                )

        return parser



from typing import cast

import click
from click.core import ParameterSource

from qualibrate_config.qulibrate_types import RawConfigType

__all__ = [
    "not_default",
    "get_config_by_args_mapping",
    "get_optional_config",
    "qualibrate_config_from_sources",
]


def not_default(ctx: click.Context, arg_key: str) -> bool:
    return ctx.get_parameter_source(arg_key) in (
        ParameterSource.COMMANDLINE,
        ParameterSource.ENVIRONMENT,
    )


def get_config_by_args_mapping(
    args_mapping: dict[str, str],
    from_file: RawConfigType,
    ctx: click.Context,
) -> RawConfigType:
    params = list(
        filter(lambda item: item[0] in args_mapping, ctx.params.items())
    )
    for arg_key, arg_value in params:
        not_default_resolver_arg = not_default(ctx, arg_key)
        if not_default_resolver_arg or args_mapping[arg_key] not in from_file:
            from_file[args_mapping[arg_key]] = arg_value
    return from_file


def get_optional_config(
    args_mapping: dict[str, str],
    from_file: RawConfigType | None,
    ctx: click.Context,
) -> RawConfigType | None:
    return get_config_by_args_mapping(args_mapping, from_file or {}, ctx)


def _get_storage_config(
    ctx: click.Context, from_file: RawConfigType
) -> RawConfigType:
    storage_keys = {
        "storage_type": "type",
        "storage_location": "location",
    }
    return get_config_by_args_mapping(storage_keys, from_file, ctx)


def _get_calibration_library_config(
    ctx: click.Context, from_file: RawConfigType | None
) -> RawConfigType | None:
    args_mapping = {
        "calibration_library_resolver": "resolver",
        "calibration_library_folder": "folder",
    }
    from_file = get_optional_config(args_mapping, from_file, ctx)
    return from_file


def _get_app_config(
    ctx: click.Context, from_file: RawConfigType
) -> RawConfigType | None:
    args_mapping = {"app_static_site_files": "static_site_files"}
    # TODO: JsonTimelineDB
    return get_optional_config(args_mapping, from_file, ctx)


def _get_runner_config(
    ctx: click.Context, from_file: RawConfigType
) -> RawConfigType | None:
    args_mapping = {"runner_address": "address", "runner_timeout": "timeout"}
    return get_optional_config(args_mapping, from_file, ctx)


def _get_composite_config(
    ctx: click.Context, from_file: RawConfigType | None
) -> RawConfigType | None:
    app = get_optional_config(
        {"spawn_app": "spawn"},
        from_file.get("app") if from_file is not None else None,
        ctx,
    )
    runner = get_optional_config(
        {"spawn_runner": "spawn"},
        from_file.get("runner") if from_file is not None else None,
        ctx,
    )
    qua_dashboards = get_optional_config(
        {"spawn_qua_dashboards": "spawn"},
        from_file.get("qua_dashboards") if from_file is not None else None,
        ctx,
    )
    if all(s is None for s in (app, runner, qua_dashboards)):
        return None
    return {
        "app": (app or {"spawn": False}),
        "runner": (runner or {"spawn": False}),
        "qua_dashboards": (qua_dashboards or {"spawn": False}),
    }


def qualibrate_config_from_sources(
    ctx: click.Context,
    from_file: RawConfigType,
    required_subconfigs: tuple[str, ...],
    optional_subconfig_names: tuple[str, ...],
) -> RawConfigType:
    for subconfig in required_subconfigs:
        if subconfig not in from_file:
            from_file[subconfig] = {}
    optional_configs_from_sources = {
        "calibration_library": _get_calibration_library_config,
        "app": _get_app_config,
        "runner": _get_runner_config,
        "composite": _get_composite_config,
    }
    optional_subconfigs = {
        name: optional_configs_from_sources[name](
            ctx, cast(RawConfigType, from_file.get(name))
        )
        for name in optional_subconfig_names
    }
    storage_config = _get_storage_config(ctx, from_file["storage"])
    qualibrate_common_mapping = {
        "password": "password",
        "version": "version",
        "project": "project",
        "log_folder": "log_folder",
    }
    qualibrate_common = get_config_by_args_mapping(
        qualibrate_common_mapping,
        from_file,
        ctx,
    )
    qualibrate_common.update(
        {
            "storage": storage_config,
            **optional_subconfigs,
        }
    )
    return qualibrate_common



from pathlib import Path

from qualibrate_config.vars import QUALIBRATE_CONFIG_KEY, QUALIBRATE_PATH


def get_user_storage() -> Path:
    return QUALIBRATE_PATH.joinpath(
        "user_storage", f"${{#/{QUALIBRATE_CONFIG_KEY}/project}}"
    )



from pathlib import Path

from qualibrate_config.models.base.config_base import BaseConfig
from qualibrate_config.models.storage_type import StorageType

__all__ = ["StorageConfig"]


class StorageConfig(BaseConfig):
    type: StorageType = StorageType.local_storage
    location: Path



import functools
import warnings

from qualibrate_config.cli.utils import deprecated_alias
from qualibrate_config.core import from_sources as _from_source_content

__all__ = _from_source_content.__all__

deprecated_m = "qualibrate_config.cli.utils.from_sources"
new_m = "qualibrate_config.core.from_sources"
warnings.warn(
    (
        f"Module '{deprecated_m}' is deprecated and will be removed in a "
        f"future version. Please use '{new_m}' instead."
    ),
    DeprecationWarning,
    stacklevel=2,
)

l_deprecated = functools.partial(
    deprecated_alias, deprecated_module=deprecated_m, new_module=new_m
)

# Apply decorator to each re-exported function
get_config_by_args_mapping = l_deprecated(
    name="get_config_by_args_mapping",
)(_from_source_content.get_config_by_args_mapping)
get_optional_config = l_deprecated(name="get_optional_config")(
    _from_source_content.get_optional_config
)
not_default = l_deprecated(name="write_config")(
    _from_source_content.not_default
)
qualibrate_config_from_sources = l_deprecated(
    name="qualibrate_config_from_sources"
)(_from_source_content.qualibrate_config_from_sources)
